from NuRadioReco.modules.base.module import register_run
from NuRadioReco.detector.ARIANNA import analog_components
import numpy as np
from NuRadioReco.utilities import units, fft
import time
import logging




class hardwareResponseIncorporator:
    """
    Incorporates the gain and phase induced by the ARIANNA hardware.


    """

    def __init__(self):
        self.logger = logging.getLogger("NuRadioReco.hardwareResponseIncorporator")
        self.__debug = False
        self.__time_delays = {}
        self.__t = 0
        self.__mingainlin = None
        self.begin()


    def begin(self, debug=False):
        self.__debug = debug
        #analog_components

    def get_filter(self, frequencies, station_id, channel_id, det, sim_to_data=False, phase_only=False, mode=None, mingainlin=None):
        """
        helper function to return the filter that the module applies.

        Parameters
        -----------
        frequencies: array of floats
            the frequency array for which the filter should be returned
        station_id: int
            the station id
        channel_id: int
            the channel id
        det: detector instance
            the detector
        the remaining arguments are the arguments of the run method!

        Returns: array of complex floats
            the complex filter amplitudes
        """
        amp_type = det.get_amplifier_type(station_id, channel_id)
        amp_measurement = det.get_amplifier_measurement(station_id, channel_id)
        amp_response = analog_components.get_amplifier_response(frequencies, amp_type=amp_type,
                                                                    amp_measurement=amp_measurement)

        if mingainlin!=None:
            mingainlin = float(mingainlin)
            ampmax= np.max(np.abs(amp_response))
            iamp_gain_low = np.where(np.abs(amp_response)<(mingainlin*ampmax))
            amp_response[iamp_gain_low]=(mingainlin*ampmax) * np.exp(1j * np.angle(amp_response[iamp_gain_low]))

        cable_response = analog_components.get_cable_response_parametrized(frequencies,
                                   *det.get_cable_type_and_length(station_id, channel_id))
        if(mode=='phase_only'):
            cable_response = np.ones_like(cable_response) * np.exp(1j * np.angle(cable_response))
            amp_response = np.ones_like(amp_response) * np.angle(amp_response)
        elif(mode=='relative'):
            ampmax= np.max(np.abs(amp_response))
            amp_response /= ampmax

        if sim_to_data:
            return amp_response * cable_response
        else:
            return 1./ (amp_response * cable_response)

    @register_run()
    def run(self, evt, station, det, sim_to_data=False, phase_only=False,mode=None,mingainlin=None):
        """
        Switch sim_to_data to go from simulation to data or otherwise.
        The option zero_noise can be used to zero the noise around the pulse. It is unclear, how useful this is.

        Parameters
        -----------
        sim_to_data: bool (default False)
            if False, deconvolve the hardware response
            if True, convolve with the hardware response
        phase_only: bool (default False)
            if True, only the phases response is applied but not the amplitude response
        mode: string
            'phase_only': only the phases response is applied but not the amplitude response
                (identical to phase_only=True )
            'relativ': gain of amp is divided by maximum of the gain, i.e. at the maximum of the
                filter response is 1 (before applying cable response). This makes it easier
                to compare the filtered to unfiltered signal
            None : default, gain and phase effects are applied 'normally'
        mingainlin: float
            In frequency ranges where the gain gets very small, the reconstruction of the original signal (obtained by
            dividing the measured signal by the gain) leads to excessively high values, due to the effect of
            post-amplifier noise. In order to mitigate this effect, a minimum gain (linear scale!) as fraction of the
            maximum gain can be defined. If specified, any gain value smaller than mingainlin will be replaced by mingainlin.
            Note: The adjustment to the minimal gain is NOT visible when getting the amp response from
            analog_components.get_amplifier_response()
        """

        self.__mingainlin = mingainlin

        if (phase_only):
            mode='phase_only'
            self.logger.warning('Please use option mode=''phase_only'' in the future, use of option phase_only will be phased out')

        t = time.time()

        for channel in station.iter_channels():
            frequencies = channel.get_frequencies()
            trace_fft = channel.get_frequency_spectrum()

            trace_fft *= self.get_filter(frequencies, station.get_id(), channel.get_id(), det, sim_to_data, phase_only, mode, mingainlin)
            # zero first bins to avoid DC offset
            trace_fft[0] = 0

            # hardwareResponse incorporator should always be used in conjunction with bandpassfilter
            # otherwise, noise will be blown up
            channel.set_frequency_spectrum(trace_fft, channel.get_sampling_rate())

            if not sim_to_data:
                # Include cable delays
                cable_delay = det.get_cable_delay(station.get_id(), channel.get_id())
                self.logger.debug("cable delay of channel {} is {}ns".format(channel.get_id(), cable_delay / units.ns))
                channel.add_trace_start_time(-cable_delay)


        self.__t += time.time() - t

    def end(self):
        from datetime import timedelta
        self.logger.setLevel(logging.INFO)
        dt = timedelta(seconds=self.__t)
        self.logger.info("total time used by this module is {}".format(dt))
        return dt

    def __calculate_time_delays_cable(self):
        """
        helper function to calculate the time delay of the amp for a delta pulse
        """
        sampling_rate = 10 * units.GHz  # assume a huge sampling rate to have a good time resolution
        n = 2 ** 12
        trace = np.zeros(n)
        trace[n // 2] = 1
        max_time = trace.argmax() / sampling_rate
        spec = fft.time2freq(trace, sampling_rate)
        ff = np.fft.rfftfreq(n, 1. / sampling_rate)
        response = analog_components.get_cable_response(ff)
        response_gain = response['gain']
        response_phase = response['phase']
        trace2 = fft.freq2time(spec * response_gain * response_phase, sampling_rate)
#         import matplotlib.pyplot as plt
#         fig, ax = plt.subplots(1, 1)
#         ax.plot(trace)
#         ax.plot(trace2)
#         plt.show()
        max_time2 = np.abs(trace2).argmax() / sampling_rate
        return max_time2 - max_time

    def __calculate_time_delays_amp(self, amp_type):
        """
        helper function to calculate the time delay of the amp for a delta pulse
        """
        amp_response_f = analog_components.amplifier_response[amp_type]
        sampling_rate = 10 * units.GHz  # assume a huge sampling rate to have a good time resolution
        n = 2 ** 12
        trace = np.zeros(n)
        trace[n // 2] = 1
        max_time = trace.argmax() / sampling_rate
        spec = fft.time2freq(trace, sampling_rate)
        ff = np.fft.rfftfreq(n, 1. / sampling_rate)
        amp_response_gain = amp_response_f['gain'](ff)
        amp_response_phase = amp_response_f['phase'](ff)
        mask = (ff < 70 * units.MHz) & (ff > 40 * units.MHz)
        spec[~mask] = 0
        trace2 = fft.freq2time(spec * amp_response_gain * amp_response_phase, sampling_rate)
#         import matplotlib.pyplot as plt
#         fig, ax = plt.subplots(1, 1)
#         ax.plot(trace)
#         ax.plot(trace2)
#         plt.show()
        max_time2 = np.abs(trace2).argmax() / sampling_rate
        return max_time2 - max_time

    def get_time_delay(self, amp_type):
        if(amp_type not in self.__time_delays.keys()):
            # not yet calculated -> calculate the time delay
            self.__time_delays[amp_type] = self.__calculate_time_delays_amp(amp_type)
            self.logger.info("time delays of amp {} have not yet been calculated -> calculating -> time delay is {:.2f} ns".format(amp_type, self.__time_delays[amp_type] / units.ns))
        return self.__time_delays[amp_type]

    def get_mingainlin(self):
        return self.__mingainlin



from NuRadioReco.modules.base.module import register_run
import numpy as np
from scipy import constants

from radiotools import helper as hp

from NuRadioReco.utilities import units, ice
from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import channelParameters as chp

import logging
logger = logging.getLogger('templateDirectionFitter')


class templateDirectionFitter:
    """
    Calculates basic signal parameters.
    """

    def __init__(self):
        pass

    def begin(self):
        pass

    @register_run()
    def run(self, evt, station, det, debug=True, channels_to_use=[0, 1, 2, 3], cosmic_ray=False):
        """
        Fits the direction using templates

        Parameters
        ----------
        evt: event

        station: station

        det: detector

        debug: bool
            set debug
        channels_to_use: list
            antenna to use for fit
        cosmic_ray: bool
            type to set correlation template

        """
        if(cosmic_ray):
            type_str = 'cr'
            xcorrelations = chp.cr_xcorrelations
        else:
            type_str = 'nu'
            xcorrelations = chp.nu_xcorrelations

        station_id = station.get_id()
        channels = station.iter_channels(channels_to_use)

        times = []
        positions = []

        for iCh, channel in enumerate(channels):
            channel_id = channel.get_id()
            times.append(channel[xcorrelations]['{}_ref_xcorr_time'.format(type_str)] + channel.get_trace_start_time())
            positions.append(det.get_relative_position(station_id, channel_id))

        times = np.array(times)
        positions = np.array(positions)
        site = det.get_site(station_id)
        n_ice = ice.get_refractive_index(-0.01, site)

        from scipy import optimize as opt

        def obj_plane(params, positions, t_measured):
            zenith, azimuth = params
            if cosmic_ray:
                if((zenith < 0) or (zenith > 0.5 * np.pi)):
                    return np.inf
            else:
                if((zenith < 0.5 * np.pi) or (zenith > np.pi)):
                    return np.inf
            v = hp.spherical_to_cartesian(zenith, azimuth)
            c = constants.c * units.m / units.s
            if not cosmic_ray:
                c = c / n_ice
                logger.debug("using speed of light = {:.4g}".format(c))
            t_expected = -(np.dot(v, positions.T) / c)
            sigma = 1 * units.ns
            chi2 = np.sum(((t_expected - t_expected.mean()) - (t_measured - t_measured.mean())) ** 2 / sigma ** 2)
            logger.debug("texp = {texp}, tm = {tmeas}, {chi2}".format(texp=t_expected, tmeas=t_measured, chi2=chi2))
            return chi2

        method = "Nelder-Mead"
        options = {'maxiter': 1000,
                   'disp': False}
        zenith_start = 135 * units.deg
        if cosmic_ray:
            zenith_start = 45 * units.deg
        starting_chi2 = {}
        for starting_az in np.array([0, 90, 180, 270]) * units.degree:
            starting_chi2[starting_az] = obj_plane((zenith_start, starting_az), positions, times)
        azimuth_start = min(starting_chi2, key=starting_chi2.get)
        res = opt.minimize(obj_plane, x0=[zenith_start, azimuth_start], args=(positions, times), method=method, options=options)

        output_str = "reconstucted angles theta = {:.1f}, phi = {:.1f}".format(res.x[0] / units.deg, hp.get_normalized_angle(res.x[1]) / units.deg)
        if station.has_sim_station():
            sim_zen = station.get_sim_station()[stnp.zenith]
            sim_az = station.get_sim_station()[stnp.azimuth]
            dOmega = hp.get_angle(hp.spherical_to_cartesian(sim_zen, sim_az), hp.spherical_to_cartesian(res.x[0], res.x[1]))
            output_str += "  MC theta = {:.1f}, phi = {:.1f},  dOmega = {:.2f}".format(sim_zen / units.deg, sim_az / units.deg, dOmega / units.deg)
        logger.info(output_str)
        station[stnp.zenith] = res.x[0]
        station[stnp.azimuth] = hp.get_normalized_angle(res.x[1])
        if(cosmic_ray):
            station[stnp.cr_zenith] = res.x[0]
            station[stnp.cr_azimuth] = hp.get_normalized_angle(res.x[1])
        else:
            station[stnp.nu_zenith] = res.x[0]
            station[stnp.nu_azimuth] = hp.get_normalized_angle(res.x[1])

    def end(self):
        pass



import dash
import json
import plotly
import numpy as np
from app import app
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from NuRadioReco.utilities import units
from NuRadioReco.framework.parameters import electricFieldParameters as efp
import NuRadioReco.eventbrowser.apps.common
import dataprovider
provider = dataprovider.DataProvider()

layout = [
    dcc.RadioItems(
        id='efield-overview-rec-sim',
        options=[
            {'label': 'Reconstruction', 'value': 'rec'},
            {'label': 'Simulation', 'value': 'sim'}
        ],
        value='rec'
    ),
    html.Div(id='efield-overview-properties')
]

efield_properties_for_overview = [
    {
        'label': 'Ray Path Type',
        'param': efp.ray_path_type,
        'unit': None
    },{
        'label': 'Zenith [deg]',
        'param': efp.zenith,
        'unit': units.deg
    },{
        'label': 'Azimuth [deg]',
        'param': efp.azimuth,
        'unit': units.deg
    },{
        'label': 'spectrum Slope',
        'param': efp.cr_spectrum_slope,
        'unit': None
    },{
        'label': 'Energy Fluence [eV]',
        'param': efp.signal_energy_fluence,
        'unit': units.eV
    },{
        'label': 'Polarization Angle [deg]',
        'param': efp.polarization_angle,
        'unit': units.deg
    },{
        'label': 'Expected Polarization Angle [deg]',
        'param': efp.polarization_angle_expectation,
        'unit': units.deg
    }
]

@app.callback(Output('efield-overview-properties', 'children'),
                [Input('filename', 'value'),
                Input('event-counter-slider', 'value'),
                Input('station-id-dropdown', 'value'),
                Input('efield-overview-rec-sim', 'value')],
                [State('user_id', 'children')])
def efield_overview_properties(filename, evt_counter, station_id, rec_sim, juser_id):
    if filename is None or station_id is None:
        return ''
    user_id = json.loads(juser_id)
    ariio = provider.get_arianna_io(user_id, filename)
    evt = ariio.get_event_i(evt_counter)
    station = evt.get_station(station_id)
    if station is None:
        return []
    reply = []
    if rec_sim == 'rec':
        chosen_station = station
    else:
        if station.get_sim_station() is None:
            return {}
        chosen_station = station.get_sim_station()
    for electric_field in chosen_station.get_electric_fields():
        props = NuRadioReco.eventbrowser.apps.common.get_properties_divs(electric_field, efield_properties_for_overview)
        reply.append(html.Div([
            html.Div('Channels', className='custom-table-td'),
            html.Div('{}'.format(electric_field.get_channel_ids()), className='custom-table-td custom-table-td-last')
        ], className='custom-table-row'))
        reply.append(html.Div(props, style={'margin': '0 0 30px'}))
    return reply



from NuRadioReco.modules.base.module import register_run
import numpy as np
from scipy import signal
import time

from NuRadioReco.utilities import units
from NuRadioReco.utilities import trace_utilities
from NuRadioReco.framework.parameters import channelParameters as chp
from NuRadioReco.framework.parameters import stationParameters as stnp

import logging
logger = logging.getLogger('channelSignalReconstructor')


class channelSignalReconstructor:
    """
    Calculates basic signal parameters.

    """

    def __init__(self):
        self.__t = 0
        self.__conversion_factor_integrated_signal = trace_utilities.conversion_factor_integrated_signal
        self.begin()

    def begin(self, debug=False, signal_start=20 * units.ns, signal_stop=100 * units.ns,
              noise_start=150 * units.ns, noise_stop=350 * units.ns):
        """
        Parameters
        -----------
        signal_start: float
            Start time (relative to the trace start time) of the window in which signal quantities will be calculated, with time units
        debug: bool
            Set module to debug output
        signal_stop: float
            Stop time (relative to the trace start time) of the window in which signal quantities will be calculated, with time units
        noise_start: float
            Start time (relative to the trace start time) of the window in which noise quantities will be calculated, with time units
        noise_stop: float
            Stop time (relative to the trace start time) of the window in which noise quantities will be calculated, with time units
        """
        self.__signal_window_start = signal_start
        self.__signal_window_stop = signal_stop
        self.__noise_window_start = noise_start
        self.__noise_window_stop = noise_stop
        self.__debug = debug

    def get_SNR(self, station_id, channel, det, stored_noise=False, rms_stage=None):
        """
        Parameters
        -----------
        channel, det
            Channel, Detector
        stored_noise: bool
            Calculates noise from pre-computed forced triggers
        rms_stage: string
            See functionality of det.get_noise_RMS

        Returns
        ----------
        SNR: dict
            dictionary of various SNR parameters
        """

        trace = channel.get_trace()
        times = channel.get_times() - channel.get_trace_start_time()

        if (self.__noise_window_start >= self.__noise_window_stop):
            logger.error("Noise cannot end before noise starts")
        if self.__signal_window_start >= self.__signal_window_stop:
            logger.error("Signal cannot end before signal starts")

        noise_window_mask = (times > self.__noise_window_start) & (times < self.__noise_window_stop)
        signal_window_mask = (times > self.__signal_window_start) & (times < self.__signal_window_stop)

        # Various definitions
        noise_int = np.sum(np.square(trace[noise_window_mask]))
        noise_int *= (self.__signal_window_stop - self.__signal_window_start) / \
            float(self.__noise_window_stop - self.__noise_window_start)

        if stored_noise:
            # we use the RMS from forced triggers
            noise_rms = det.get_noise_RMS(station_id, channel.get_id(), stage=rms_stage)
        else:
            noise_rms = np.sqrt(np.mean(np.square(trace[noise_window_mask])))

        if self.__debug:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(times[signal_window_mask], np.square(trace[signal_window_mask]))
            plt.plot(times[noise_window_mask], np.square(trace[noise_window_mask]), c='k', label='noise')
            plt.xlabel("Times [ns]")
            plt.ylabel("Power")
            plt.legend()

        # Calculating SNR
        SNR = {}
        if (noise_rms == 0) or (noise_int == 0):
            logger.info("RMS of noise is zero, calculating an SNR is not useful. All SNRs are set to infinity.")
            SNR['peak_2_peak_amplitude'] = np.infty
            SNR['peak_amplitude'] = np.infty
            SNR['integrated_power'] = np.infty
        else:

            SNR['integrated_power'] = (np.sum(np.square(trace[signal_window_mask])) - noise_int)
            if SNR['integrated_power'] < noise_int:
                logger.debug("Integrated signal {0} smaller than noise {1}, power SNR 0".format(SNR['integrated_power'], noise_int))
                SNR['integrated_power'] = 0.
            else:

                SNR['integrated_power'] /= (noise_int / self.__signal_window_start)
                SNR['integrated_power'] = np.sqrt(SNR['integrated_power'])

            # calculate amplitude values
            SNR['peak_2_peak_amplitude'] = np.max(trace[signal_window_mask]) - np.min(trace[signal_window_mask])
            SNR['peak_2_peak_amplitude'] /= noise_rms
            SNR['peak_2_peak_amplitude'] /= 2

            SNR['peak_amplitude'] = np.max(np.abs(trace[signal_window_mask])) / noise_rms

        # SCNR
        SNR['Seckel_2_noise'] = 5

        if self.__debug:
            plt.figure()
            plt.plot(times, trace)
            plt.axvline(self.__noise_window_start, c='k', label='Noise Window')
            plt.axvline(self.__noise_window_stop, c='k', linestyle='--')
            plt.axvline(self.__signal_window_start, c='r', label='Signal Window')
            plt.axvline(self.__signal_window_stop, c='r', linestyle='--')
            plt.legend()
            plt.show()

        return SNR

    @register_run()
    def run(self, evt, station, det, stored_noise=False, rms_stage='amp'):
        """
        Parameters
        -----------
        evt, station, det
            Event, Station, Detector
        stored_noise: bool
            Calculates noise from pre-computed forced triggers
        rms_stage: string
            See functionality of det.get_noise_RMS
        """

        t = time.time()
        max_amplitude_station = 0
        for channel in station.iter_channels():
            times = channel.get_times()
            trace = channel.get_trace()
            h = np.abs(signal.hilbert(trace))
            max_amplitude = np.max(np.abs(trace))
            channel[chp.signal_time] = times[np.argmax(h)]
            max_amplitude_station = max(max_amplitude_station, max_amplitude)
            channel[chp.maximum_amplitude] = max_amplitude
            channel[chp.maximum_amplitude_envelope] = h.max()
            channel[chp.P2P_amplitude] = np.max(trace) - np.min(trace)

            # Use noise precalculated from forced triggers
            channel[chp.SNR] = self.get_SNR(station.get_id(), channel, det,
                                            stored_noise=stored_noise, rms_stage=rms_stage)

        station[stnp.channels_max_amplitude] = max_amplitude

        self.__t = time.time() - t

    def end(self):
        from datetime import timedelta
        logger.setLevel(logging.INFO)
        dt = timedelta(seconds=self.__t)
        logger.info("total time used by this module is {}".format(dt))
        return dt



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['NuRadioReco',
 'NuRadioReco.detector',
 'NuRadioReco.detector.ARA',
 'NuRadioReco.detector.ARIANNA',
 'NuRadioReco.detector.RNO_G',
 'NuRadioReco.eventbrowser',
 'NuRadioReco.eventbrowser.apps',
 'NuRadioReco.eventbrowser.apps.trace_plots',
 'NuRadioReco.framework',
 'NuRadioReco.modules',
 'NuRadioReco.modules.ARA',
 'NuRadioReco.modules.ARIANNA',
 'NuRadioReco.modules.RNO_G',
 'NuRadioReco.modules.base',
 'NuRadioReco.modules.custom',
 'NuRadioReco.modules.custom.deltaT',
 'NuRadioReco.modules.io',
 'NuRadioReco.modules.io.araroot',
 'NuRadioReco.modules.io.coreas',
 'NuRadioReco.modules.io.noise',
 'NuRadioReco.modules.io.snowshovel',
 'NuRadioReco.modules.phasedarray',
 'NuRadioReco.modules.trigger',
 'NuRadioReco.utilities']

package_data = \
{'': ['*'],
 'NuRadioReco': ['autodoc/*',
                 'autodoc/custom_scripts/*',
                 'autodoc/pages/*',
                 'examples/*',
                 'examples/AraData/*',
                 'examples/EnvelopePhasedArray/*',
                 'examples/EnvelopePhasedArray/Effective_volume/*',
                 'examples/EnvelopePhasedArray/Noise_trigger_rate/*',
                 'examples/PhasedArray/*',
                 'examples/PhasedArray/Effective_volume/*',
                 'examples/PhasedArray/Noise_trigger_rate/*',
                 'examples/PhasedArray/SNR_curves/*',
                 'examples/PhasedArray/SNR_curves/results/*',
                 'examples/StandAloneScripts/*',
                 'examples/example_data/*',
                 'test/*',
                 'test/tiny_reconstruction/*',
                 'test/trigger_tests/*'],
 'NuRadioReco.detector': ['amps/*',
                          'amps/NTU01/*',
                          'filter/*',
                          'filter/SBLP-467+_S2P/*',
                          'filter/SHP-20+_S2P/*',
                          'filter/SHP-25+_S2P/*',
                          'filter/SHP-48+_S2P/*',
                          'filter/SXHP-48+_S2P/*',
                          'filter/ZFHP-1R2-S+_S2P/*',
                          'filter/ZFLP-450+_S2P/*',
                          'filter/ZX75HP-44+_S2P/*',
                          'filter/ZX75LP-470+_S2P/*',
                          'test/*',
                          'visualization/*',
                          'visualization/assets/*'],
 'NuRadioReco.detector.ARA': ['HardwareResponses/*'],
 'NuRadioReco.detector.ARIANNA': ['HardwareResponses/*'],
 'NuRadioReco.detector.RNO_G': ['HardwareResponses/*'],
 'NuRadioReco.eventbrowser': ['assets/*',
                              'assets/fontawesome/*',
                              'assets/fontawesome/css/*',
                              'assets/fontawesome/font/*'],
 'NuRadioReco.eventbrowser.apps': ['cosmic_ray_plots/*',
                                   'overview_plots/*',
                                   'simulation_plots/*'],
 'NuRadioReco.modules': ['envelope_phasedarray/*']}

install_requires = \
['tinydb',
 'tinydb-serialization',
 'aenum',
 'astropy<=2.0.14',
 'radiotools >=0.2.0',
 'h5py',
 'peakutils',
 'requests']

extras_require = \
{'doc': ['sphinx']}

setup(name='NuRadioReco',
      version='1.1.1',
      description='NuRadioReco: A reconstruction framework for radio neutrino detectors',
      author='Christian Glaser',
      author_email='work@c-glaser.de',
      url='https://github.com/nu-radio/NuRadioReco',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>3.4',
     )



"""
This file generates the input files for the electron neutrino effective
volume simulation  using a phased array. For creating a test file, run:

python T01generate_event_list.py

For generating a set of energies organised in folders for later running
on a cluster with the appropriate scripts, run:

python T01generate_event_list.py full

WARNING: This file needs NuRadioMC installed. https://github.com/nu-radio/NuRadioMC
"""

from __future__ import absolute_import, division, print_function
from NuRadioReco.utilities import units
from NuRadioMC.EvtGen.generator import generate_eventlist_cylinder
import numpy as np
import os
import sys

if ( len(sys.argv) < 2 ):
	mode = 'simple'
else:
	mode = sys.argv[1]

# define simulation volume
zmin = -2.7 * units.km  # the ice sheet at South Pole is 2.7km deep
zmax = 0 * units.km
rmin = 0 * units.km
rmax = 4 * units.km

costhetas = np.linspace(1, -1, 2) # Usually 20
thetas = np.arccos(costhetas)
#thetas = np.linspace(0.*units.deg, 180.*units.deg, 3)
thetamins = thetas[0:-1]
thetamaxs = thetas[1:]
print(thetamins)
print(thetamaxs)

phimin = 0.*units.deg
phimax = 360.*units.deg

if ( mode == 'full' ):
	logEs = np.linspace(15., 20., 50) # Usually 50
else:
	logEs = np.linspace(16., 19, 2)
Es = 10**logEs * units.eV
Emins = Es[0:-1]
Emaxs = Es[1:]

flavours = [12] # +/-12: electronic neutrino. +/-14: muonic neutrino. +/-16: tau neutrino
if ( mode == 'full' ):
	nevt = 1e6
	nevt_perfile = 1e5
elif mode == 'minimal':
	nevt = 1e2
	nevt_perfile = 1e3
else:
	nevt = 1e4
	nevt_perfile = 1e4

for thetamin, thetamax in zip(thetamins, thetamaxs):

	for flavour in flavours:

		folder_angle = "{:.2f}rad_{}".format(thetamin,flavour)
		try:
			os.mkdir(folder_angle)
		except:
			pass

		folderlist = []
		for ifolder, Emin,Emax in zip(range(len(Emins)),Emins,Emaxs):
			folderlist.append("{:.2f}_{}_{}_{:.2e}_{:.2e}".format(thetamin,flavour,str(ifolder).zfill(2),Emin,Emax))
		for folder in folderlist:
			try:
				os.mkdir(os.path.join(folder_angle, folder))
			except:
				pass

		for folder, Emin, Emax in zip(folderlist, Emins, Emaxs):

			input_dir = os.path.join(folder_angle, folder, 'input')
			try:
				os.mkdir(input_dir)
			except:
				pass
			if mode == 'simple':
				outname = folder+'.hdf5'
			elif mode == 'minimal':
				outname = 'minimal_eventlist.hdf5'
			else:
				outname = folder_angle+'/'+folder+'/input/'+folder+'.hdf5'
			print(Emin/units.PeV, Emax/units.PeV)
			print(outname)
			generate_eventlist_cylinder(outname, nevt, Emin, Emax, rmin, rmax, zmin, zmax, full_rmin=rmin, full_rmax=rmax, full_zmin=zmin, full_zmax=zmax, thetamin=thetamin, thetamax=thetamax, phimin=phimin, phimax=phimax, start_event_id=1, flavor=[flavour], n_events_per_file=nevt_perfile, deposited=True)

# generate one event list at 1e19 eV with 100 neutrinos
#generate_eventlist_cylinder('1e19_n1e2.hdf5', 1e2, 1e19 * units.eV, 1e19 * units.eV, rmin, rmax, zmin, zmax)

# generate one event list at 1e19 eV with 1000 neutrinos
#generate_eventlist_cylinder('1e19_n1e3.hdf5', 1e3, 1e19 * units.eV, 1e19 * units.eV, rmin, rmax, zmin, zmax)

# generate one event list at 1e18 eV with 10000 neutrinos
#generate_eventlist_cylinder('1e18_n1e4.hdf5', 1e4, 1e18 * units.eV, 1e18 * units.eV, rmin, rmax, zmin, zmax)



import pickle
import NuRadioReco.framework.hybrid_shower

class HybridInformation():
    def __init__(self):
        self.__hybrid_showers = []
        self.__hybrid_shower_names = []

    def add_hybrid_shower(self, hybrid_shower):
        if hybrid_shower.get_name() in self.__hybrid_shower_names:
            raise ValueError('A hybrid shower with the name {} already exists'.format(hybrid_shower.get_name()))
        self.__hybrid_showers.append(hybrid_shower)
        self.__hybrid_shower_names.append(hybrid_shower.get_name())

    def get_hybrid_showers(self):
        for shower in self.__hybrid_showers:
            yield shower

    def get_hybrid_shower(self, name):
        if name not in self.__hybrid_shower_names:
            raise ValueError('Could not find hybrid shower with name {}'.format(name))
        return self.__hybrid_showers[self.__hybrid_shower_names.index(name)]

    def serialize(self):
        shower_pickles = []
        for shower in self.get_hybrid_showers():
            shower_pickles.append(shower.serialize())
        data = {
            'shower_pickles': shower_pickles
        }
        return pickle.dumps(data, protocol=4)

    def deserialize(self, data_pkl):
        for shower_pkl in pickle.loads(data_pkl)['shower_pickles']:
            shower = NuRadioReco.framework.hybrid_shower.HybridShower('')
            shower.deserialize(shower_pkl)
            self.add_hybrid_shower(shower)



from NuRadioReco.modules.base.module import register_run
import numpy as np
from numpy.polynomial import polynomial as poly
from scipy import signal
from scipy.signal import correlate
from scipy import optimize as opt
import matplotlib.pyplot as plt
import time

from radiotools import helper as hp
from radiotools import plthelpers as php

from NuRadioReco.utilities import trace_utilities

from NuRadioReco.utilities import geometryUtilities as geo_utl
from NuRadioReco.utilities import units
from NuRadioReco.detector import antennapattern
from NuRadioReco.modules.voltageToEfieldConverter import get_array_of_channels
import NuRadioReco.framework.base_trace

from NuRadioReco.framework.parameters import stationParameters as stnp
from NuRadioReco.framework.parameters import electricFieldParameters as efp

from NuRadioReco.framework import electric_field as ef

import logging



class voltageToEfieldConverterPerChannel:
    """
    Converts voltage trace to electric field per channel

    Assumes either co-pole or cross-pole, then finds the electric field per channel by E = V/H in fourier space.
    """

    def __init__(self):
        self.logger = logging.getLogger('NuRadioReco.voltageToEfieldConverterPerChannel')
        self.__counter = 0
        self.begin()

    def begin(self):
        self.antenna_provider = antennapattern.AntennaPatternProvider()

    @register_run()
    def run(self, evt, station, det, pol=0, debug=True):
        """
        Performs computation for voltage trace to electric field per channel

        Will provide a deconvoluted (electric field) trace for each channel from the stations input voltage traces

        Paramters
        ---------
        evt: event data structure
            the event data structure
        station: station data structure
            the station data structure
        det: detector object
            the detector object
        pol: polarization
            0 = eTheta polarized, 1 = ePhi polarized
        debug: bool
            if True additional debug plot(s) are displayed - currently unused
        """
        self.__counter += 1
        event_time = station.get_station_time()
        station_id = station.get_id()
        self.logger.debug("event {}, station {}".format(evt.get_id(), station_id))
        if station.get_sim_station() is not None and station.get_sim_station().has_parameter(stnp.zenith):
            zenith = station.get_sim_station()[stnp.zenith]
            azimuth = station.get_sim_station()[stnp.azimuth]
        else:
            self.logger.debug("Using reconstructed angles as no simulation present")
            zenith = station[stnp.zenith]
            azimuth = station[stnp.azimuth]

        frequencies = station.get_channel(0).get_frequencies()  # assuming that all channels have the  same sampling rate and length
        use_channels = det.get_channel_ids(station.get_id())
        efield_antenna_factor = trace_utilities.get_efield_antenna_factor(station, frequencies, use_channels, det,
                                                                          zenith, azimuth, self.antenna_provider)

        sampling_rate = station.get_channel(0).get_sampling_rate()

        for iCh, channel in enumerate(station.iter_channels()):
            efield = ef.ElectricField([iCh])
            trace = channel.get_frequency_spectrum()
            mask1 = np.abs(efield_antenna_factor[iCh][0]) != 0
            mask2 = np.abs(efield_antenna_factor[iCh][1]) != 0
            efield_spectrum = np.zeros((3, len(trace)), dtype=np.complex)
            efield_spectrum[1][mask1] = (1.0 - pol) ** 2 * trace[mask1] / efield_antenna_factor[iCh][0][mask1]
            efield_spectrum[2][mask2] = pol ** 2 * trace[mask2] / efield_antenna_factor[iCh][1][mask2]
            efield.set_frequency_spectrum(efield_spectrum, sampling_rate)
            efield.set_trace_start_time(channel.get_trace_start_time())
            efield[efp.zenith] = zenith
            efield[efp.azimuth] = azimuth
            station.add_electric_field(efield)

    def end(self):
        pass



from NuRadioReco.modules.base.module import register_run
from NuRadioReco.utilities import units
from NuRadioReco.framework.trigger import SimplePhasedTrigger
from NuRadioReco.modules.analogToDigitalConverter import analogToDigitalConverter
import numpy as np
from scipy import constants
import time
import logging
logger = logging.getLogger('phasedTriggerSimulator')

cspeed = constants.c * units.m / units.s

main_low_angle = -53. * units.deg
main_high_angle = 47. * units.deg
default_angles = np.arcsin(np.linspace(np.sin(main_low_angle), np.sin(main_high_angle), 15))

default_sec_angles = 0.5 * (default_angles[:-1] + default_angles[1:])
default_sec_angles = np.insert(default_sec_angles, len(default_sec_angles), default_angles[-1] + 3.5 * units.deg)


class triggerSimulator:
    """
    Calculates the trigger for a phased array with a primary and a secondary beam.
    The channels that participate in both beams and the pointing angle for each
    subbeam can be specified.

    See https://arxiv.org/pdf/1809.04573.pdf
    """

    def __init__(self):
        self.__t = 0
        self.begin()

    def begin(self, debug=False, pre_trigger_time=100 * units.ns):
        self.__pre_trigger_time = pre_trigger_time
        self.__debug = debug

    def get_antenna_positions(self, station, det, triggered_channels=None, component=2):
        """
        Calculates the vertical coordinates of the antennas of the detector
        """

        ant_pos = [ det.get_relative_position(station.get_id(), channel.get_id())[component] \
                    for channel in station.iter_channels() \
                    if channel.get_id() in triggered_channels ]

        return np.array(ant_pos)

    def get_beam_rolls(self, station, det, triggered_channels,
                       phasing_angles=default_angles, ref_index=1.55,
                       trigger_adc=False):
        """
        Calculates the delays needed for phasing the array.
        """
        station_id = station.get_id()
        sampling_rate = None
        value_error = False

        for channel in station.iter_channels(use_channels=triggered_channels):
            channel_id = channel.get_id()
            if trigger_adc:
                if sampling_rate is None:
                    sampling_rate = det.get_channel(station_id, channel_id)["trigger_adc_sampling_frequency"]
                elif sampling_rate != det.get_channel(station_id, channel_id)["trigger_adc_sampling_frequency"]:
                    value_error = True
            else:
                if sampling_rate is None:
                        sampling_rate = channel.get_sampling_rate()
                elif sampling_rate != channel.get_sampling_rate():
                    value_error = True
            if value_error:
                error_msg  = 'Phased array channels do not have matching sampling rates. '
                error_msg += 'Please specify a common sampling rate.'
                raise ValueError(error_msg)

        time_step = 1. / sampling_rate

        ant_z = self.get_antenna_positions(station, det, triggered_channels, 2)
        self.check_vertical_string(station, det, triggered_channels)
        beam_rolls = []
        ref_z = (np.max(ant_z) + np.min(ant_z)) / 2

        for angle in phasing_angles:
            subbeam_rolls = {}
            for z, channel_id in zip(ant_z, triggered_channels):
                delay = (z - ref_z) / cspeed * ref_index * np.sin(angle)
                roll = int(delay / time_step)
                subbeam_rolls[channel_id] = roll
            logger.debug("angle:", angle / units.deg)
            logger.debug(subbeam_rolls)
            beam_rolls.append(subbeam_rolls)

        return beam_rolls

    def get_channel_trace_start_time(self, station, triggered_channels):

        channel_trace_start_time = None
        for channel in station.iter_channels(use_channels=triggered_channels):
            if channel_trace_start_time is None:
                channel_trace_start_time = channel.get_trace_start_time()
            elif channel_trace_start_time != channel.get_trace_start_time():
                error_msg  = 'Phased array channels do not have matching trace start times. '
                error_msg += 'This module is not prepared for this case.'
                raise ValueError(error_msg)

        return channel_trace_start_time

    def check_vertical_string(self, station, det, triggered_channels):
        """
        Checks if the triggering antennas lie in a straight vertical line
        """

        cut = 1.e-3 * units.m
        ant_x = self.get_antenna_positions(station, det, triggered_channels, 0)
        diff_x = np.abs(ant_x - ant_x[0])
        ant_y = self.get_antenna_positions(station, det, triggered_channels, 1)
        diff_y = np.abs(ant_y - ant_y[0])
        if (sum(diff_x) > cut or sum(diff_y) > cut):
            raise NotImplementedError('The phased triggering array should lie on a vertical line')

    def phased_trigger(self,
                       station,
                       det,
                       beam_rolls,
                       sec_beam_rolls,
                       triggered_channels,
                       threshold,
                       window_time=10.67 * units.ns,
                       cut_times=(None,None),
                       trigger_adc=False):
        """
        Calculates the trigger for a certain phasing configuration.
        Beams are formed. A set of overlapping time windows is created and
        the square of the voltage is averaged within each window.
        If the average surpasses the given threshold squared times the
        number of antennas in any of these windows, a trigger is created.
        This is an ARA-like power trigger.

        Parameters
        ----------
        station: Station object
            Description of the current station
        det: Detector object
            Description of the current detector
        beam_rolls: array of ints
            Contains the integers for rolling the voltage traces (delays)
        sec_beam_rolls: array of ints
            Contains the secondary beam integers for rolling the voltage traces (delays)
        triggered_channels: array of ints
            Ids of the triggering channels
        threshold: float
            Voltage threshold for a SINGLE antenna. It will be rescaled with the
            square root of the number of antennas.
        window_time: float
            Width of the time window used in the power integration
        cut_times: (float, float) tuple
            Times for cutting the trace. This helps reducing the number of noise-induced triggers.
        trigger_adc: bool
            If True, analog to digital conversion is performed. It must be specified in the
            detector file. See analogToDigitalConverter module for information

        Returns
        -------
        is_triggered: bool
            True if the triggering condition is met
        trigger_delays: dictionary
            the delays for the primary channels that have caused a trigger.
            If there is no trigger, it's an empty dictionary
        sec_trigger_delays: dictionary
            the delays for the secondary channels that have caused a trigger.
            If there is no trigger or no secondary channels, it's an empty dictionary
        """
        station_id = station.get_id()

        traces = {}
        for channel in station.iter_channels(use_channels=triggered_channels):
            channel_id = channel.get_id()
            time_step = 1 / channel.get_sampling_rate()

            if trigger_adc:

                ADC = analogToDigitalConverter()
                trace = ADC.get_digital_trace(station, det, channel,
                                        trigger_adc=trigger_adc,
                                        random_clock_offset=True,
                                        adc_type='perfect_floor_comparator')
                time_step = 1 / det.get_channel(station_id, channel_id)['trigger_adc_sampling_frequency']
                times  = np.arange(len(trace), dtype=np.float)
                times += channel.get_trace_start_time()

            else:

                trace = np.copy(channel.get_trace())  # get the enveloped trace
                times = np.copy(channel.get_times())  # get the corresponding time bins

            if cut_times != (None,None):
                left_bin = np.argmin(np.abs(times-cut_times[0]))
                right_bin = np.argmin(np.abs(times-cut_times[1]))
                trace[0:left_bin] = 0
                trace[right_bin:None] = 0

            traces[channel_id] = trace[:]

        for subbeam_rolls, sec_subbeam_rolls in zip(beam_rolls, sec_beam_rolls):

            phased_trace = None
            # Number of antennas: primary beam antennas + secondary beam antennas
            Nant = len(beam_rolls[0]) + len(sec_beam_rolls[0])

            for channel_id in traces:

                trace = traces[channel_id]

                if(phased_trace is None):
                    phased_trace = np.roll(trace, subbeam_rolls[channel_id])
                else:
                    phased_trace += np.roll(trace, subbeam_rolls[channel_id])

                if(channel_id in sec_subbeam_rolls):
                    phased_trace += np.roll(trace, sec_subbeam_rolls[channel_id])

            # Implmentation of the ARA-like power trigger
            window_width = int(window_time / time_step)
            n_windows = int(len(trace) / window_width)
            n_windows = 2 * n_windows - 1
            squared_mean_threshold = Nant * threshold ** 2

            # Create a sliding window
            strides = phased_trace.strides
            windowed_traces = np.lib.stride_tricks.as_strided(phased_trace, \
                              shape=(n_windows, window_width), \
                              strides=(int(window_width / 2) * strides[0], strides[0]))

            squared_mean = np.sum(windowed_traces ** 2 / window_width, axis=1)

            if True in (squared_mean > squared_mean_threshold):
                trigger_delays = {}
                for channel_id in subbeam_rolls:
                    trigger_delays[channel_id] = subbeam_rolls[channel_id] * time_step
                sec_trigger_delays = {}
                for channel_id in sec_subbeam_rolls:
                    sec_trigger_delays[channel_id] = sec_subbeam_rolls[channel_id] * time_step
                logger.debug("Station has triggered")
                return True, trigger_delays, sec_trigger_delays

        return False, {}, {}

    @register_run()
    def run(self, evt, station, det,
            threshold=60 * units.mV,
            triggered_channels=None,
            secondary_channels=None,
            trigger_name='simple_phased_threshold',
            phasing_angles=default_angles,
            secondary_phasing_angles=default_sec_angles,
            set_not_triggered=False,
            window_time=10.67 * units.ns,
            coupled=True,
            ref_index=1.75,
            cut_times=(None,None),
            trigger_adc=False):
        """
        simulates phased array trigger for each event

        Several channels are phased by delaying their signals by an amount given
        by a pointing angle. Several pointing angles are possible in order to cover
        the sky. The array triggered_channels controls the channels that are phased,
        according to the angles phasing_angles. A secondary phasing that is added
        to this primary phasing is possible, and it is controlled by the parametres
        secondary_channels and secondary_phasing_angles.

        Parameters
        ----------
        evt: Event object
            Description of the current event
        station: Station object
            Description of the current station
        det: Detector object
            Description of the current detector
        threshold: float
            threshold above (or below) a trigger is issued, absolute amplitude
        triggered_channels: array of ints
            channels ids of the channels that form the primary phasing array
            if None, all channels are taken
        secondary_channels: array of int
            channel ids of the channels that form the secondary phasing array
            if None, no channels are taken
        trigger_name: string
            name for the trigger
        phasing_angles: array of float
            pointing angles for the primary beam
        secondary_phasing_angles: array of float
            pointing angles for the secondary beam
        set_not_triggered: bool (default: False)
            if True not trigger simulation will be performed and this trigger will be set to not_triggered
        window_time: float
            Width of the time window used in the power integration
        coupled: bool
            if True, the primary sub-beams are paired to the secondary sub-beams.
            if False, the primary sub-beams and the secondary sub-beams specify independent beams.
        ref_index: float
            refractive index for beam forming
        cut_times: (float, float) tuple
            Times for cutting the trace. This helps reducing the number of noise-induced triggers.
        trigger_adc: bool
            If True, analog to digital conversion is performed. It must be specified in the
            detector file. See analogToDigitalConverter module for information

        Returns
        -------
        is_triggered: bool
            True if the triggering condition is met
        """

        if (triggered_channels == None):
        	triggered_channels = [channel._id for channel in station.iter_channels()]

        if (secondary_channels == None):
            # If there are no chosen secondary channels, their list is empty
            secondary_channels = []

        if set_not_triggered:

            is_triggered = False
            trigger_delays = {}
            sec_trigger_delays = {}

        else:

            channel_trace_start_time = self.get_channel_trace_start_time(station, triggered_channels)

            logger.debug("primary channels:", triggered_channels)
            beam_rolls = self.get_beam_rolls(station, det, triggered_channels, phasing_angles,
                                             ref_index=ref_index, trigger_adc=trigger_adc)
            empty_rolls = [ {} for direction in range(len(phasing_angles)) ]
            logger.debug("secondary_channels:", secondary_channels)
            if (len(secondary_channels) == 0):
                only_primary = True
            else:
                only_primary = False
                secondary_beam_rolls = self.get_beam_rolls(station, det, secondary_channels, secondary_phasing_angles,
                                                           ref_index=ref_index, trigger_adc=trigger_adc)

            if only_primary:
                is_triggered, trigger_delays, sec_trigger_delays = self.phased_trigger(station, det,
                        beam_rolls, empty_rolls, triggered_channels, threshold, window_time, cut_times,
                        trigger_adc=trigger_adc)
            elif coupled:
                is_triggered, trigger_delays, sec_trigger_delays = self.phased_trigger(station, det,
                        beam_rolls, secondary_beam_rolls, triggered_channels, threshold, window_time, cut_times,
                        trigger_adc=trigger_adc)
            else:
                primary_trigger, trigger_delays, dummy_delays = self.phased_trigger(station, det, beam_rolls, empty_rolls,
                        triggered_channels, threshold, window_time, cut_times, trigger_adc=trigger_adc)
                secondary_trigger, sec_trigger_delays, dummy_delays = self.phased_trigger(station, det, secondary_beam_rolls, empty_rolls,
                        secondary_channels, threshold, window_time, cut_times, trigger_adc=trigger_adc)
                is_triggered = primary_trigger or secondary_trigger

        trigger = SimplePhasedTrigger(trigger_name, threshold, triggered_channels, secondary_channels,
                                      phasing_angles, secondary_phasing_angles,
                                      trigger_delays, sec_trigger_delays)
        trigger.set_triggered(is_triggered)
        if is_triggered:
            trigger.set_trigger_time(channel_trace_start_time)
        else:
            trigger.set_trigger_time(None)
        station.set_trigger(trigger)

        return is_triggered

    def end(self):
        pass



import logging
import time
import numpy as np
import NuRadioReco.modules.channelResampler
from NuRadioReco.utilities import units
from scipy.interpolate import interp1d
from scipy.signal import resample
from NuRadioReco.modules.base.module import register_run

def perfect_comparator(trace, adc_n_bits, adc_ref_voltage, mode='floor', output='voltage'):
    """
    Simulates a perfect comparator flash ADC that compares the voltage to the
    voltage for the least significative bit and takes the floor or the ceiling
    of their ratio as a digitised value of the trace.

    Parameters
    ----------
    trace: array of floats
        Trace containing the voltage to be digitised
    adc_n_bits: int
        Number of bits of the ADC
    adc_ref_voltage: float
        Voltage corresponding to the maximum number of counts given by the
        ADC: 2**adc_n_bits - 1
    mode: string
        'floor' or 'ceiling'
    output: string
        - 'voltage' to store the ADC output as discretised voltage trace
        - 'counts' to store the ADC output in ADC counts

    Returns
    -------
    digital_trace: array of floats
        Digitised voltage trace in volts or ADC counts
    """

    lsb_voltage = adc_ref_voltage/(2 ** (adc_n_bits-1) - 1)

    if (mode == 'floor'):
        digital_trace = np.floor(trace / lsb_voltage)
    elif (mode == 'ceiling'):
        digital_trace = np.ceil(trace / lsb_voltage)
    else:
        raise ValueError('Choose floor or ceiing as modes for the comparator ADC')

    digital_trace = apply_saturation(digital_trace, adc_n_bits, adc_ref_voltage)
    digital_trace = round_to_int(digital_trace)

    if (output == 'voltage'):
        digital_trace = lsb_voltage * digital_trace.astype(np.float)
    elif (output == 'counts'):
        pass
    else:
        raise ValueError("The ADC output format is unknown. Please choose 'voltage' or 'counts'" )

    return digital_trace

def perfect_floor_comparator(trace, adc_n_bits, adc_ref_voltage, output='voltage'):
    """
    Perfect comparator ADC that takes the floor value of the comparison.
    See perfect_comparator
    """

    return perfect_comparator(trace, adc_n_bits, adc_ref_voltage, mode='floor', output=output)

def perfect_ceiling_comparator(trace, adc_n_bits, adc_ref_voltage, output='voltage'):
    """
    Perfect comparator ADC that takes the floor value of the comparison.
    See perfect_floor.
    """

    return perfect_comparator(trace, adc_n_bits, adc_ref_voltage, mode='ceiling', output=output)

def apply_saturation(adc_counts_trace, adc_n_bits, adc_ref_voltage):
    """
    Takes a digitised trace in ADC counts and clips the parts of the
    trace with values higher than 2**(adc_n_bits-1)-1 or lower than
    -2**(adc_n_bits-1).

    Parameters
    ----------
    adc_counts_trace: array of floats
        Voltage in ADC counts, unclipped
    adc_n_bits: int
        Number of bits of the ADC
    adc_ref_voltage: float
        Voltage corresponding to the maximum number of counts given by the
        ADC: 2**(adc_n_bits-1) - 1

    Returns
    -------
    saturated_trace: array of floats
        The clipped or saturated voltage trace
    """

    saturated_trace = adc_counts_trace[:]

    highest_count = 2 ** (adc_n_bits-1) - 1
    high_saturation_mask = adc_counts_trace > highest_count
    saturated_trace[high_saturation_mask] = highest_count

    lowest_count = - 2 ** (adc_n_bits-1)
    low_saturation_mask = adc_counts_trace < lowest_count
    saturated_trace[low_saturation_mask] = lowest_count

    return saturated_trace

def round_to_int(digital_trace):

    int_trace = np.rint(digital_trace)
    int_trace = int_trace.astype(int)

    return int_trace

class analogToDigitalConverter():
    """
    This class simulates an analog to digital converter. The steps followed
    by this module to achieve the conversion are:
    1) The following properties of the channel are read. They must be in the
    detector configuration file:
        - "adc_nbits", the number of bits of the ADC
        - "adc_reference_voltage", the reference voltage in volts, that is, the
        maximum voltage the ADC can convert without saturating
        - "adc_sampling_frequency", the sampling frequency in GHz
    2) A random clock offset (jitter) can be added, as it would happen in
    a real experiment. Choose random_clock_offset = True to do so. A time
    delay can also be fixed if the field "adc_time_delay" is specified in ns.
    The channel trace is interpolated to get the trace values at the clock
    times displaced from the channel times. This is fine as long as the input
    channel traces have been simulated with a sampling rate greater than the
    ADC sampling rate, which should be the case.
    3) A type of ADC converter is chosen, which transforms the trace in ADC
    counts (discrete values). The available types are listed in the list
    _adc_types, which are (see functions with the same names for documentation):
        - 'perfect_floor_comparator'
        - 'perfect_ceiling_comparator'

    IMPORTANT: Since this module already performs a downsampling, there is no
    need to use the channelResampler in those channels that possess an ADC.

    Note that after this module the traces are still expressed in voltage units,
    only the possible values are discretised.

    If the ADC is used for triggering and the user does not want to modify the
    trace, the function get_digital_trace can be used. If there are two different
    ADCs for the same channel, one for triggering and another one for storing,
    one can define a trigger ADC adding "trigger_" to every relevant ADC field
    in the detector configuration, and use them setting trigger_adc to True in
    get_digital_trace.
    """

    def __init__(self):
        self.__t = 0
        self._adc_types = {'perfect_floor_comparator': perfect_floor_comparator,
                           'perfect_ceiling_comparator': perfect_ceiling_comparator}
        self._mandatory_fields = ['adc_nbits',
                                  'adc_reference_voltage',
                                  'adc_sampling_frequency']

        self.logger = logging.getLogger('NuRadioReco.analogToDigitalConverter')

    def get_digital_trace(self, station, det, channel,
                          trigger_adc=False,
                          random_clock_offset=True,
                          adc_type='perfect_floor_comparator',
                          diode=None,
                          return_sampling_frequency=False,
                          output='voltage'):
        """
        Returns the digital trace for a channel, without setting it. This allows
        the creation of a digital trace that can be used for triggering purposes
        without removing the original information on the channel.

        Parameters
        ----------
        station: framework.station.Station object
        det: detector.detector.Detector object
        channel: framework.channel.Channel object
        trigger_adc: bool
            If True, the relevant ADC parameters in the config file are the ones
            that start with 'trigger_'
        random_clock_offset: bool
            If True, a random clock offset between -1 and 1 clock cycles is added
        adc_type: string
            The type of ADC used. The following are available:
            - perfect_floor_comparator
            - perfect_ceiling_comparator
            See functions with the same name on this module for documentation
        diode: utilities.diodeSimulator.diodeSimulator object
            Diode used to envelope filter the signal
        return_sampling_frequency: bool
            If True, returns the trace and the ADC sampling frequency
        output: string
            - 'voltage' to store the ADC output as discretised voltage trace
            - 'counts' to store the ADC output in ADC counts

        Returns
        -------
        digital_trace: array of floats
            Digitised voltage trace
        adc_sampling_frequency: float
            ADC sampling frequency for the channel
        """

        station_id = station.get_id()
        channel_id = channel.get_id()
        det_channel = det.get_channel(station_id, channel_id)

        for field in self._mandatory_fields:
            if trigger_adc:
                field_check = 'trigger_' + field
            else:
                field_check = field
            if field_check not in det_channel:
                channel_id = channel.get_id()
                error_msg  = "The field {} is not present in channel {}. ".format(field_check,
                                                                                  channel_id)
                error_msg += "Please specify it on your detector file"
                raise ValueError(error_msg)

        times = channel.get_times()[:]
        trace = channel.get_trace()[:]
        if diode is not None:
            trace = diode.tunnel_diode(channel)

        if trigger_adc:
            adc_time_delay_label = "trigger_adc_time_delay"
            adc_n_bits_label = "trigger_adc_nbits"
            adc_ref_voltage_label = "trigger_adc_reference_voltage"
            adc_sampling_frequency_label = "trigger_adc_sampling_frequency"
        else:
            adc_time_delay_label = "adc_time_delay"
            adc_n_bits_label = "adc_nbits"
            adc_ref_voltage_label = "adc_reference_voltage"
            adc_sampling_frequency_label = "adc_sampling_frequency"

        adc_time_delay = 0

        if adc_time_delay_label in det_channel:
            if det_channel[adc_time_delay_label] is not None:
                adc_time_delay = channel[adc_time_delay_label] * units.ns

        if random_clock_offset:
            clock_offset = np.random.uniform(-1, 1)
            adc_time_delay += clock_offset / channel.get_sampling_rate()

        adc_n_bits = det_channel[adc_n_bits_label]
        adc_ref_voltage = det_channel[adc_ref_voltage_label] * units.V
        adc_sampling_frequency = det_channel[adc_sampling_frequency_label] * units.GHz

        if (adc_sampling_frequency > channel.get_sampling_rate()):
            error_msg  = 'The ADC sampling rate is greater than '
            error_msg += 'the channel {} sampling rate. '.format(channel.get_id())
            error_msg += 'Please change the ADC sampling rate.'
            raise ValueError(error_msg)

        delayed_times = times + adc_time_delay
        interpolate_trace = interp1d(times, trace, kind='quadratic',
                                     fill_value='extrapolate')

        delayed_trace = interpolate_trace(delayed_times)

        new_n_samples = int( (adc_sampling_frequency / channel.get_sampling_rate()) * len(delayed_trace) )
        resampled_trace = resample(delayed_trace, new_n_samples)

        digital_trace = self._adc_types[adc_type](delayed_trace, adc_n_bits,
                                                  adc_ref_voltage, output)

        if return_sampling_frequency:
            return digital_trace, adc_sampling_frequency
        else:
            return digital_trace

    @register_run()
    def run(self, evt, station, det,
            random_clock_offset=True,
            adc_type='perfect_floor_comparator',
            output='voltage'):
        """
        Runs the analogToDigitalConverter and transforms the traces from all
        the channels of an input station to digital voltage values.

        Parameters
        ----------
        evt: framework.event.Event object
        station: framework.station.Station object
        det: detector.detector.Detector object
        random_clock_offset: bool
            If True, a random clock offset between -1 and 1 clock cycles is added
        adc_type: string
            The type of ADC used. The following are available:
            - perfect_floor_comparator
            See functions with the same name on this module for documentation
        output: string
            - 'voltage' to store the ADC output as discretised voltage trace
            - 'counts' to store the ADC output in ADC counts
        """

        t = time.time()

        for channel in station.iter_channels():

            digital_trace, adc_sampling_frequency = self.get_digital_trace(station, det, channel,
                                                        random_clock_offset=random_clock_offset,
                                                        adc_type=adc_type,
                                                        return_sampling_frequency=True,
                                                        output=output)

            channel.set_trace(digital_trace, adc_sampling_frequency)

        self.__t += time.time() - t


    def end(self):
        from datetime import timedelta
        self.logger.setLevel(logging.INFO)
        dt = timedelta(seconds=self.__t)
        logger.info("total time used by this module is {}".format(dt))
        return dt



from __future__ import absolute_import, division, print_function  # , unicode_literals
import dash_html_components as html
import plotly
from NuRadioReco.utilities import units
import NuRadioReco.eventbrowser.apps.trace_plots.rec_electric_field_trace
import NuRadioReco.eventbrowser.apps.trace_plots.rec_electric_field_spectrum
import NuRadioReco.eventbrowser.apps.trace_plots.channel_time_trace
import NuRadioReco.eventbrowser.apps.trace_plots.channel_spectrum
import NuRadioReco.eventbrowser.apps.trace_plots.multi_channel_plot
import numpy as np
import logging

logger = logging.getLogger('traces')

layout = html.Div([
    html.Div(id='trigger-trace', style={'display': 'none'}),
    html.Div([
        html.Div([
            html.Div('Electric Field Traces', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.trace_plots.rec_electric_field_trace.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex': '1'}),
        html.Div([
            html.Div('Electric Field Spectrum', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.trace_plots.rec_electric_field_spectrum.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex': '1'})
    ], style={'display': 'flex'}),
    html.Div([
        html.Div([
            html.Div('Channel Traces', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.trace_plots.channel_time_trace.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex': '1'}),
        html.Div([
            html.Div('Channel Spectrum', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.trace_plots.channel_spectrum.layout,
            className='panel-body')
        ], className='panel panel-default', style={'flex': '1'})
    ], style={'display': 'flex'}),
    html.Div([
        html.Div([
            html.Div('Individual Channels', className='panel-heading'),
            html.Div(NuRadioReco.eventbrowser.apps.trace_plots.multi_channel_plot.layout, className='panel-body')
        ], className='panel panel-default', style={'flex': '1'})
    ], style={'display': 'flex'})
])



from __future__ import absolute_import, division, print_function, unicode_literals
import NuRadioReco.framework.event
import NuRadioReco.detector.detector
import NuRadioReco.detector.generic_detector
import NuRadioReco.modules.io.event_parser_factory
import numpy as np
import logging
import pickle
import time

VERSION = 2
VERSION_MINOR = 2


class NuRadioRecoio(object):

    def __init__(self, filenames, parse_header=True, parse_detector=True, fail_on_version_mismatch=True,
                 fail_on_minor_version_mismatch=False,
                 max_open_files=10, log_level=None):
        """
        Initialize NuRadioReco io

        Parameters
        ----------
        filenames: string or list of strings
            the input file/files
        parse_header: boolean
            If True, the event headers are parsed and can be accessed through
            the get_header() function
        parse_detector: boolean
            If True, detector information in the files is parsed and can be
            accessed through the get_detector() function
        fail_on_version_mismatch: boolean
            Controls if the module should try to read files with a different major version
        fail_on_minor_version_mismatch: boolean
            Controls if the module should try to read files with a different minor version
        max_open_files: int
            the maximum number of files that remain open simultaneously
        """
        if(not isinstance(filenames, list)):
            filenames = [filenames]
        self.__file_scanned = False
        self.logger = logging.getLogger('NuRadioReco.NuRadioRecoio')
        self.logger.info("initializing NuRadioRecoio with file {}".format(filenames))
        t = time.time()
        if log_level is not None:
            self.logger.setLevel(log_level)
        self.__fail_on_version_mismatch = fail_on_version_mismatch
        self.__fail_on_minor_version_mismatch = fail_on_minor_version_mismatch
        self.__parse_header = parse_header
        self.__parse_detector = parse_detector
        self.__read_lock = False
        self.__max_open_files = max_open_files
        self.openFile(filenames)
        self._current_file_id = 0
        self.logger.info("... finished in {:.0f} seconds".format(time.time() - t))

    def _get_file(self, iF):
        if(iF not in self.__open_files):
            self.logger.debug("file {} is not yet open, opening file".format(iF))
            self.__open_files[iF] = {}
            self.__open_files[iF]['file'] = open(self._filenames[iF], 'rb')
            self.__open_files[iF]['time'] = time.time()
            self.__check_file_version(iF)
            if(len(self.__open_files) > self.__max_open_files):
                self.logger.debug("more than {} file are open, closing oldest file".format(self.__max_open_files))
                tnow = time.time()
                iF_close = 0
                for key, value in self.__open_files.items():
                    if(value['time'] < tnow):
                        tnow = value['time']
                        iF_close = key
                self.logger.debug("closing file {} that was opened at {}".format(iF_close, tnow))
                self.__open_files[iF_close]['file'].close()
                del self.__open_files[iF_close]
        return self.__open_files[iF]['file']

    def __check_file_version(self, iF):
        self.__file_version = int.from_bytes(self._get_file(iF).read(6), 'little')
        self.__file_version_minor = int.from_bytes(self._get_file(iF).read(6), 'little')
        if(self.__file_version != VERSION):
            self.logger.error("data file not readable. File has version {}.{} but current version is {}.{}".format(self.__file_version, self.__file_version_minor,
                                                                                                              VERSION, VERSION_MINOR))
            if(self.__fail_on_version_mismatch):
                raise IOError
        if(self.__file_version_minor != VERSION_MINOR):
            self.logger.error("data file might not readable. File has version {}.{} but current version is {}.{}".format(self.__file_version, self.__file_version_minor,
                                                                                                                    VERSION, VERSION_MINOR))
            if(self.__fail_on_minor_version_mismatch):
                raise IOError
        self.__scan_files = NuRadioReco.modules.io.event_parser_factory.scan_files_function(self.__file_version, self.__file_version_minor)
        self.__iter_events = NuRadioReco.modules.io.event_parser_factory.iter_events_function(self.__file_version, self.__file_version_minor)

    def openFile(self, filenames):
        self._filenames = filenames
        self.__n_events = 0
        self.__event_ids = []
        self._bytes_start_header = [[]]
        self._bytes_length_header = [[]]
        self._bytes_start = [[]]
        self._bytes_length = [[]]
        self.__open_files = {}
        self._detector_dicts = {}
        self.__detectors = {}
        self._event_specific_detector_changes = {}

        self.__event_headers = {}
        if(self.__parse_header):
            self.__scan_files()

    def close_files(self):
        for f in self.__open_files.values():
            f['file'].close()

    def get_filenames(self):
        return self._filenames

    def _parse_event_header(self, evt_header):
        self.__event_ids.append(evt_header['event_id'])
        for station_id, station in evt_header['stations'].items():
            if station_id not in self.__event_headers:
                self.__event_headers[station_id] = {}
            for key, value in station.items():
                # treat sim_station differently
                if(key == 'sim_station'):
                    pass
                else:
                    if key not in self.__event_headers[station_id]:
                        self.__event_headers[station_id][key] = []
                    self.__event_headers[station_id][key].append(value)

    def __scan_files(self):
        current_byte = 12  # skip datafile header
        iF = 0
        while True:
            self._get_file(iF).seek(current_byte)
            continue_loop, iF, current_byte = self.__scan_files(self, iF, current_byte)
            if not continue_loop:
                break

        self.__event_ids = np.array(self.__event_ids)
        self.__file_scanned = True

        # compute number of events
        n = 0
        for b in self._bytes_start:
            n += len(b)
        self.__n_events = n

        # convert lists to numpy arrays for convenience
        for station_id, station in self.__event_headers.items():
            for key, value in station.items():
                self.__event_headers[station_id][key] = np.array(value)

    def get_header(self):
        if(not self.__file_scanned):
            self.__scan_files()
        return self.__event_headers

    def get_event_ids(self):
        """
        returns a list of (run, eventid) tuples of all events contained in the data file
        """
        if(not self.__file_scanned):
            self.__scan_files()
        return self.__event_ids

    def get_event_i(self, event_number):
        while(self.__read_lock):
            time.sleep(1)
            self.logger.debug("read lock waiting 1ms")
        self.__read_lock = True

        if(not self.__file_scanned):
            self.__scan_files()
        if(event_number < 0 or event_number >= self.get_n_events()):
            self.logger.error('event number {} out of bounds, only {} present in file'.format(event_number, self.get_n_events()))
            return None
        # determine in which file event i is
        istart = 0
        file_id = 0
        for iF in range(len(self._filenames)):
            istop = istart + len(self._bytes_start[iF])
            if((event_number >= istart) and (event_number < istop)):
                file_id = iF
                event_id = event_number - istart
                break
            else:
                istart = istop

        self._get_file(file_id).seek(self._bytes_start[file_id][event_id])
        evtstr = self._get_file(file_id).read(self._bytes_length[file_id][event_id])
        event = NuRadioReco.framework.event.Event(0, 0)
        event.deserialize(evtstr)
        self.__read_lock = False
        self._current_file_id = file_id
        self._current_event_id = event.get_id()
        self._current_run_number = event.get_run_number()
        # # If the event file contains a detector description that is a
        # # generic detector, it might have event-specific properties and we
        # # need to set the detector to the right event
        if self._current_file_id in self.__detectors.keys():
            if 'generic_detector' in self._detector_dicts[self._current_file_id]:
                if self._detector_dicts[self._current_file_id]['generic_detector']:
                    self.__detectors[self._current_file_id].set_event(self._current_run_number, self._current_event_id)
        return event

    def get_event(self, event_id):
        if(not self.__file_scanned):
            self.__scan_files()
        mask = (self.__event_ids[:, 0] == event_id[0]) & (self.__event_ids[:, 1] == event_id[1])
        if(np.sum(mask) == 0):
            self.logger.error('event number {} not found in file'.format(event_id))
            return None
        elif(np.sum(mask) > 1):
            self.logger.warning(f"{np.sum(mask):d} events with the same run event id pair found. Returning first occurence.")
        self._current_run_number = event_id[0]
        self._current_event_id = event_id[1]
        i = np.argwhere(mask)[0][0]
        return self.get_event_i(i)

    def get_events(self):
        self._current_file_id = 0
        self._get_file(self._current_file_id).seek(12)  # skip file header
        for event in self.__iter_events(self):
            self._current_event_id = event.get_id()
            self._current_run_number = event.get_run_number()
            if self._current_file_id in self.__detectors.keys():
                if 'generic_detector' in self._detector_dicts[self._current_file_id]:
                    if self._detector_dicts[self._current_file_id]['generic_detector']:
                        self.__detectors[self._current_file_id].set_event(self._current_run_number, self._current_event_id)
            yield event

    def get_detector(self):
        """
        If parse_detector was set True in the __init__() function, this function return
        the detector description (assuming there is one in the files). If several
        files with different detectors are read, the detector for the last returned
        event is given.
        """
        # Check if detector object for current file already exists
        if self._current_file_id not in self.__detectors.keys():
            # Detector object for current file does not exist, so we create it
            if self._current_file_id not in self._detector_dicts:
                self.__scan_files()  # Maybe we just forgot to scan the file
                if self._current_file_id not in self._detector_dicts:
                    raise AttributeError('The current file does not contain a detector description.')
            detector_dict = self._detector_dicts[self._current_file_id]
            if 'generic_detector' in detector_dict.keys():
                if detector_dict['generic_detector']:
                    # Detector is a generic detector, so we have to consider default
                    # station/channel and event-specific changes
                    self.__detectors[self._current_file_id] = NuRadioReco.detector.generic_detector.GenericDetector.__new__(NuRadioReco.detector.generic_detector.GenericDetector)
                    self.__detectors[self._current_file_id].__init__(source='dictionary', json_filename='', dictionary=detector_dict, default_station=detector_dict['default_station'], default_channel=detector_dict['default_channel'])
                    if self._current_file_id in self._event_specific_detector_changes.keys():
                        for change in self._event_specific_detector_changes[self._current_file_id]:
                            self.__detectors[self._current_file_id].add_station_properties_for_event(
                                properties=change['properties'],
                                station_id=change['station_id'],
                                run_number=change['run_number'],
                                event_id=change['event_id']
                            )
                    self.__detectors[self._current_file_id].set_event(self._current_run_number, self._current_event_id)
                    return self.__detectors[self._current_file_id]
            # Detector is a normal detector
            self.__detectors[self._current_file_id] = NuRadioReco.detector.detector.Detector.__new__(NuRadioReco.detector.detector.Detector)
            self.__detectors[self._current_file_id].__init__(source='dictionary', json_filename='', dictionary=self._detector_dicts[self._current_file_id])
        # Detector object for current file already exists. If it is a generic detector,
        # we update it to the run number and ID of the last event that was requested
        # (in case there are event-specific changes) and return  it
        if 'generic_detector' in self._detector_dicts[self._current_file_id].keys():
            if self._detector_dicts[self._current_file_id]['generic_detector']:
                    self.__detectors[self._current_file_id].set_event(self._current_run_number, self._current_event_id)
        return self.__detectors[self._current_file_id]

    def get_n_events(self):
        if(not self.__file_scanned):
            self.__scan_files()
        return self.__n_events



import numpy as np
from NuRadioReco.utilities import units
from radiotools import helper as hp
import os
import inspect
import logging
from tinydb import TinyDB, Query
from tinydb_serialization import SerializationMiddleware
from tinydb.storages import MemoryStorage
import astropy.time
from datetime import datetime
from tinydb_serialization import Serializer
import six  # # used for compatibility between py2 and py3
logger = logging.getLogger('NuRadioReco.detector')


class DateTimeSerializer(Serializer):
    """
    helper class to serialize datetime objects with TinyDB
    """
    OBJ_CLASS = datetime  # The class this serializer handles

    def encode(self, obj):
        return obj.strftime('%Y-%m-%dT%H:%M:%S')

    def decode(self, s):
        return datetime.strptime(s, '%Y-%m-%dT%H:%M:%S')


serialization = SerializationMiddleware()
serialization.register_serializer(DateTimeSerializer(), 'TinyDate')


def buffer_db(in_memory, filename=None):
    """
    buffers the complet SQL database into a TinyDB object (either in memory or into a local JSON file)
    """
    logger.info("buffering SQL database on-the-fly")
    db = None
    if(in_memory):
        db = TinyDB(storage=MemoryStorage)
    else:
        db = TinyDB(filename, storage=serialization, sort_keys=True, indent=4, separators=(',', ': '))
    db.purge()

    from NuRadioReco.detector import detector_sql
    sqldet = detector_sql.Detector()
    results = sqldet.get_everything_stations()
    table_stations = db.table('stations')
    table_stations.purge()
    for result in results:
        table_stations.insert({'station_id': result['st.station_id'],
                               'commission_time': result['st.commission_time'],
                               'decommission_time': result['st.decommission_time'],
                               'station_type': result['st.station_type'],
                               'position': result['st.position'],
                               'board_number': result['st.board_number'],
                               'MAC_address': result['st.MAC_address'],
                               'MBED_type': result['st.MBED_type'],
                               'pos_position': result['pos.position'],
                               'pos_measurement_time': result['pos.measurement_time'],
                               'pos_easting': result['pos.easting'],
                               'pos_northing': result['pos.northing'],
                               'pos_altitude': result['pos.altitude'],
                               'pos_zone': result['pos.zone'],
                               'pos_site': result['pos.site']})

    table_channels = db.table('channels')
    table_channels.purge()
    results = sqldet.get_everything_channels()
    for channel in results:
        table_channels.insert({ 'station_id': channel['st.station_id'],
                                'channel_id': channel['ch.channel_id'],
                                'commission_time': channel['ch.commission_time'],
                                'decommission_time': channel['ch.decommission_time'],
                                'ant_type': channel['ant.antenna_type'],
                                'ant_orientation_phi': channel['ant.orientation_phi'],
                                'ant_orientation_theta': channel['ant.orientation_theta'],
                                'ant_rotation_phi': channel['ant.rotation_phi'],
                                'ant_rotation_theta': channel['ant.rotation_theta'],
                                'ant_position_x': channel['ant.position_x'],
                                'ant_position_y': channel['ant.position_y'],
                                'ant_position_z': channel['ant.position_z'],
                                'ant_deployment_time': channel['ant.deployment_time'],
                                'ant_comment': channel['ant.comment'],
                                'cab_length': channel['cab.cable_length'],
                                'cab_reference_measurement': channel['cab.reference_measurement'],
                                'cab_time_delay': channel['cab.time_delay'],
                                'cab_id': channel['cab.cable_id'],
                                'cab_type': channel['cab.cable_type'],
                                'amp_type': channel['amps.amp_type'],
                                'amp_reference_measurement': channel['amps.reference_measurement'],
                                'adc_id': channel['adcs.adc_id'],
                                'adc_time_delay': channel['adcs.time_delay'],
                                'adc_nbits': channel['adcs.nbits'],
                                'adc_n_samples': channel['adcs.n_samples'],
                                'adc_sampling_frequency': channel['adcs.sampling_frequency']})

    results = sqldet.get_everything_positions()
    table_positions = db.table('positions')
    table_positions.purge()
    for result in results:
        table_positions.insert({
                               'pos_position': result['pos.position'],
                               'pos_measurement_time': result['pos.measurement_time'],
                               'pos_easting': result['pos.easting'],
                               'pos_northing': result['pos.northing'],
                               'pos_altitude': result['pos.altitude'],
                               'pos_zone': result['pos.zone'],
                               'pos_site': result['pos.site']})

    logger.info("sql database buffered")
    return db


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if Singleton._instances.get(cls, None) is None:
            Singleton._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return Singleton._instances[cls]


@six.add_metaclass(Singleton)
class Detector(object):
    """
    main detector class which provides access to the detector description

    This class provides functions for all relevant detector properties.
    """

    def __init__(self, source='json', json_filename='ARIANNA/arianna_detector_db.json',
                 dictionary=None, assume_inf=True):
        """
        Initialize the stations detector properties.

        Parameters
        ----------
        source : str
            'json', 'dictionary' or 'sql'
            default value is 'json'
            if dictionary is specified, the dictionary passed to __init__ is used
            if 'sql' is specified, the file 'detector_sql_auth.json' file needs to be present in this folder that
            specifies the sql server credentials (see 'detector_sql_auth.json.sample' for an example of the syntax)
        json_filename : str
            the path to the json detector description file (if first checks a path relative to this directory, then a
            path relative to the current working directory of the user)
            default value is 'ARIANNA/arianna_detector_db.json'
        assume_inf : Bool
            Default to True, if true forces antenna madels to have infinite boundary conditions, otherwise the antenna madel will be determined by the station geometry.
        """
        if(source == 'sql'):
            self._db = buffer_db(in_memory=True)
        elif source == 'dictionary':
            self._db = TinyDB(storage=MemoryStorage)
            self._db.purge()
            stations_table = self._db.table('stations', cache_size=1000)
            for station in dictionary['stations'].values():
                stations_table.insert(station)
            channels_table = self._db.table('channels', cache_size=1000)
            for channel in dictionary['channels'].values():
                channels_table.insert(channel)
        else:
            dir_path = os.path.dirname(os.path.realpath(__file__))  # get the directory of this file
            filename = os.path.join(dir_path, json_filename)
            if(not os.path.exists(filename)):
                # try local folder instead
                filename2 = json_filename
                if(not os.path.exists(filename2)):
                    logger.error("can't locate json database file {} or {}".format(filename, filename2))
                    raise NameError
                filename = filename2
            logger.warning("loading detector description from {}".format(os.path.abspath(filename)))
            self._db = TinyDB(filename, storage=serialization,
                               sort_keys=True, indent=4, separators=(',', ': '))

        self._stations = self._db.table('stations', cache_size=1000)
        self._channels = self._db.table('channels', cache_size=1000)
        self.__positions = self._db.table('positions', cache_size=1000)

        logger.info("database initialized")

        self._buffered_stations = {}
        self.__buffered_positions = {}
        self._buffered_channels = {}
        self.__valid_t0 = astropy.time.Time('2100-1-1')
        self.__valid_t1 = astropy.time.Time('1970-1-1')

        self.__noise_RMS = None

        self.__current_time = None

        self.__assume_inf = assume_inf

    def __query_channel(self, station_id, channel_id):
        Channel = Query()
        if self.__current_time is None:
            raise ValueError("Detector time is not set. The detector time has to be set using the Detector.update() function before it can be used.")
        res = self._channels.get((Channel.station_id == station_id) & (Channel.channel_id == channel_id)
                                           & (Channel.commission_time <= self.__current_time.datetime)
                                           & (Channel.decommission_time > self.__current_time.datetime))
        if(res is None):
            logger.error("query for station {} and channel {} at time {} returned no results".format(station_id, channel_id, self.__current_time))
            raise LookupError
        return res

    def _query_channels(self, station_id):
        Channel = Query()
        if self.__current_time is None:
            raise ValueError("Detector time is not set. The detector time has to be set using the Detector.update() function before it can be used.")
        return self._channels.search((Channel.station_id == station_id)
                                           & (Channel.commission_time <= self.__current_time.datetime)
                                           & (Channel.decommission_time > self.__current_time.datetime))

    def _query_station(self, station_id):
        Station = Query()
        if self.__current_time is None:
            raise ValueError("Detector time is not set. The detector time has to be set using the Detector.update() function before it can be used.")
        res = self._stations.get((Station.station_id == station_id)
                                       & (Station.commission_time <= self.__current_time.datetime)
                                       & (Station.decommission_time > self.__current_time.datetime))
        if(res is None):
            logger.error("query for station {} at time {} returned no results".format(station_id, self.__current_time.datetime))
            raise LookupError("query for station {} at time {} returned no results".format(station_id, self.__current_time.datetime))
        return res

    def __query_position(self, position_id):
        Position = Query()
        res = self.__positions.get((Position.pos_position == position_id))
        if self.__current_time is None:
            raise ValueError("Detector time is not set. The detector time has to be set using the Detector.update() function before it can be used.")
        if(res is None):
            logger.error("query for position {} at time {} returned no results".format(position_id, self.__current_time.datetime))
            raise LookupError("query for position {} at time {} returned no results".format(position_id, self.__current_time.datetime))
        return res

    def get_station_ids(self):
        """
        returns a sorted list of all station ids present in the database
        """
        station_ids = []
        res = self._stations.all()
        if(res is None):
            logger.error("query for stations returned no results")
            raise LookupError("query for stations returned no results")
        for a in res:
            if(a['station_id'] not in station_ids):
                station_ids.append(a['station_id'])
        return sorted(station_ids)

    def _get_station(self, station_id):
        if(station_id not in self._buffered_stations.keys()):
            self._buffer(station_id)
        return self._buffered_stations[station_id]

    def get_station(self, station_id):
        return self._get_station(station_id)

    def __get_position(self, position_id):
        if(position_id not in self.__buffered_positions.keys()):
            self.__buffer_position(position_id)
        return self.__buffered_positions[position_id]

    def __get_channels(self, station_id):
        if(station_id not in self._buffered_stations.keys()):
            self._buffer(station_id)
        return self._buffered_channels[station_id]

    def __get_channel(self, station_id, channel_id):
        if(station_id not in self._buffered_stations.keys()):
            self._buffer(station_id)
        return self._buffered_channels[station_id][channel_id]

    def _buffer(self, station_id):
        self._buffered_stations[station_id] = self._query_station(station_id)
        self.__valid_t0 = astropy.time.Time(self._buffered_stations[station_id]['commission_time'])
        self.__valid_t1 = astropy.time.Time(self._buffered_stations[station_id]['decommission_time'])
        channels = self._query_channels(station_id)
        self._buffered_channels[station_id] = {}
        for channel in channels:
            self._buffered_channels[station_id][channel['channel_id']] = channel
            self.__valid_t0 = max(self.__valid_t0, astropy.time.Time(channel['commission_time']))
            self.__valid_t1 = min(self.__valid_t1, astropy.time.Time(channel['decommission_time']))

    def __buffer_position(self, position_id):
        self.__buffered_positions[position_id] = self.__query_position(position_id)

    def __get_t0_t1(self, station_id):
        Station = Query()
        res = self._stations.get(Station.station_id == station_id)
        t0 = None
        t1 = None
        if(isinstance(res, list)):
            for station in res:
                if(t0 is None):
                    t0 = station['commission_time']
                else:
                    t0 = min(t0, station['commission_time'])
                if(t1 is None):
                    t1 = station['decommission_time']
                else:
                    t1 = max(t1, station['decommission_time'])
        else:
            t0 = res['commission_time']
            t1 = res['decommission_time']
        return astropy.time.Time(t0), astropy.time.Time(t1)

    def has_station(self, station_id):
        """
        checks if a station is present in the database

        Parameters
        ----------
        station_id: int
            the station id

        Returns bool
        """
        Station = Query()
        res = self._stations.get(Station.station_id == station_id)
        return res != None

    def get_unique_time_periods(self, station_id):
        """
        returns the time periods in which the station configuration (including all channels) was constant

        Parameters
        ----------
        station_id: int
            the station id

        Returns datetime tuple
        """
        up = []
        t0, t1 = self.__get_t0_t1(station_id)
        self.update(t0)
        while True:
            if(len(up) > 0 and up[-1] == t1):
                break
            self._buffer(station_id)
            if(len(up) == 0):
                up.append(self.__valid_t0)
            up.append(self.__valid_t1)
            self.update(self.__valid_t1)
        return up

    def update(self, time):
        """
        updates the detector description to a new time

        Parameters
        ----------
        timestamp: astropy.time.Time
            the time to update the detectordescription to
            for backward compatibility datetime is also accepted, but astropy.time is prefered
        """
        if isinstance(time, datetime):
            self.__current_time = astropy.time.Time(time)
        else:
            self.__current_time = time
        logger.info("updating detector time to {}".format(self.__current_time))
        if(not ((self.__current_time > self.__valid_t0) and (self.__current_time < self.__valid_t1))):
            self._buffered_stations = {}
            self._buffered_channels = {}
            self.__valid_t0 = astropy.time.Time('2100-1-1')
            self.__valid_t1 = astropy.time.Time('1970-1-1')

    def get_channel(self, station_id, channel_id):
        """
        returns a dictionary of all channel parameters

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns: dict of channel parameters
        """
        return self.__get_channel(station_id, channel_id)

    def get_absolute_position(self, station_id):
        """
        get the absolute position of a specific station

        Parameters
        ---------
        station_id: int
            the station id

        Returns: 3-dim array of absolute station position in easting, northing and depth wrt. to snow level at
        time of measurement
        """
        res = self._get_station(station_id)
        easting, northing, altitude = 0, 0, 0
        unit_xy = units.m
        if('pos_zone' in res and res['pos_zone'] == "SP-grid"):
            unit_xy = units.feet
        if(res['pos_easting'] is not None):
            easting = res['pos_easting'] * unit_xy
        if(res['pos_northing'] is not None):
            northing = res['pos_northing'] * unit_xy
        if(res['pos_altitude'] is not None):
            altitude = res['pos_altitude']
        return np.array([easting, northing, altitude])

    def get_absolute_position_site(self, site):
        """
        get the absolute position of a specific station

        Parameters
        ---------
        site: string
            the position identifier e.g. "G"

        Returns: 3-dim array of absolute station position in easting, northing and depth wrt. to snow level at
        time of measurement
        """
        res = self.__get_position(site)
        unit_xy = units.m
        if('pos_zone' in res and res['pos_zone'] == "SP-grid"):
            unit_xy = units.feet
        easting, northing, altitude = 0, 0, 0
        if(res['pos_easting'] is not None):
            easting = res['pos_easting'] * unit_xy
        if(res['pos_northing'] is not None):
            northing = res['pos_northing'] * unit_xy
        if(res['pos_altitude'] is not None):
            altitude = res['pos_altitude'] * units.m
        return np.array([easting, northing, altitude])

    def get_relative_position(self, station_id, channel_id):
        """
        get the relative position of a specific channels/antennas with respeect to the station center

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns: 3-dim array of relative station position
        """
        res = self.__get_channel(station_id, channel_id)
        return np.array([res['ant_position_x'], res['ant_position_y'], res['ant_position_z']])

    def get_relative_positions(self, station_id):
        """
        get the relative positions of all channels/antennas with respeect to the station center

        Parameters
        ---------
        station_id: int
            the station id

        Returns: List of 3-dim array of relative station positions
        """
        res = self.__get_channels(station_id)
        positions = np.zeros((len(res), 3))
        for i, r in enumerate(res.values()):
            logger.debug("position channel {}: {:.0f}m, {:.0f}m, {:.0f}m".format(r['channel_id'], r['ant_position_x'], r['ant_position_y'], r['ant_position_z']))
            positions[i] = [r['ant_position_x'], r['ant_position_y'], r['ant_position_z']]
        return positions

    def get_site(self, station_id):
        """
        get the site where the station is deployed (e.g. MooresBay or South Pole)

        Parameters
        ---------
        station_id: int
            the station id

        Returns string
        """

        res = self._get_station(station_id)
        return res['pos_site']

    def get_number_of_channels(self, station_id):
        """
        Get the number of channels per statoin

        Parameters
        ---------
        station_id: int
            the station id

        Returns int
        """
        res = self.__get_channels(station_id)
        return len(res)

    def get_channel_ids(self, station_id):
        """
        get the channel ids of a station

        Parameters
        ---------
        station_id: int
            the station id

        Returns list of ints
        """
        channel_ids = []
        for channel in self.__get_channels(station_id).values():
            channel_ids.append(channel['channel_id'])
        return channel_ids

    def get_parallel_channels(self, station_id):
        """
        get a list of parallel antennas

        Parameters
        ---------
        station_id: int
            the station id

        Returns list of list of ints
        """
        res = self.__get_channels(station_id)
        orientations = np.zeros((len(res), 4))
        antenna_types = []
        channel_ids = []
        for iCh, ch in enumerate(res.values()):
            channel_id = ch['channel_id']
            channel_ids.append(channel_id)
            antenna_types.append(self.get_antenna_type(station_id, channel_id))
            orientations[iCh] = self.get_antenna_orientation(station_id, channel_id)
            orientations[iCh][3] = hp.get_normalized_angle(orientations[iCh][3], interval=np.deg2rad([0, 180]))
        channel_ids = np.array(channel_ids)
        antenna_types = np.array(antenna_types)
        orientations = np.round(np.rad2deg(orientations))  # round to one degree to overcome rounding errors
        parallel_antennas = []
        for antenna_type in np.unique(antenna_types):
            for u_zen_ori in np.unique(orientations[:, 0]):
                for u_az_ori in np.unique(orientations[:, 1]):
                    for u_zen_rot in np.unique(orientations[:, 2]):
                        for u_az_rot in np.unique(orientations[:, 3]):
                            mask = (antenna_types == antenna_type) \
                                     & (orientations[:, 0] == u_zen_ori) & (orientations[:, 1] == u_az_ori) \
                                     & (orientations[:, 2] == u_zen_rot) & (orientations[:, 3] == u_az_rot)
                            if(np.sum(mask)):
                                parallel_antennas.append(channel_ids[mask])
        return np.array(parallel_antennas)

    def get_cable_delay(self, station_id, channel_id):
        """
        returns the cable delay of a channel

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns float (delay time)
        """
        res = self.__get_channel(station_id, channel_id)
        return res['cab_time_delay']

    def get_cable_type_and_length(self, station_id, channel_id):
        """
        returns the cable type (e.g. LMR240) and its length

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns typle (string, float)
        """
        res = self.__get_channel(station_id, channel_id)
        return res['cab_type'], res['cab_length'] * units.m

    def get_antenna_type(self, station_id, channel_id):
        """
        returns the antenna type

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns string
        """
        res = self.__get_channel(station_id, channel_id)
        return res['ant_type']

    def get_antenna_deployment_time(self, station_id, channel_id):
        """
        returns the time of antenna deployment

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns datetime
        """
        res = self.__get_channel(station_id, channel_id)
        return res['ant_deployment_time']

    def get_antenna_orientation(self, station_id, channel_id):
        """
        returns the orientation of a specific antenna

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns typle of floats
            * orientation theta: boresight direction (zenith angle, 0deg is the zenith, 180deg is straight down)
            * orientation phi: boresight direction (azimuth angle counting from East counterclockwise)
            * rotation theta: rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector
            * rotation phi: rotation of the antenna, is perpendicular to 'orientation', for LPDAs: vector in plane of tines pointing away from connector
        """
        res = self.__get_channel(station_id, channel_id)
        return np.deg2rad([res['ant_orientation_theta'], res['ant_orientation_phi'], res['ant_rotation_theta'], res['ant_rotation_phi']])

    def get_amplifier_type(self, station_id, channel_id):
        """
        returns the type of the amplifier

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns string
        """
        res = self.__get_channel(station_id, channel_id)
        return res['amp_type']

    def get_amplifier_measurement(self, station_id, channel_id):
        """
        returns a unique reference to the amplifier measurement

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns string
        """
        res = self.__get_channel(station_id, channel_id)
        return res['amp_reference_measurement']

    def get_sampling_frequency(self, station_id, channel_id):
        """
        returns the sampling frequency

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns float
        """
        res = self.__get_channel(station_id, channel_id)
        return res['adc_sampling_frequency'] * units.GHz

    def get_number_of_samples(self, station_id, channel_id):
        """
        returns the number of samples of a channel

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id

        Returns int
        """
        res = self.__get_channel(station_id, channel_id)
        return res['adc_n_samples']

    def get_antenna_model(self, station_id, channel_id, zenith=None):
        """
        determines the correct antenna model from antenna type, position and orientation of antenna

        so far only infinite firn and infinite air cases are differentiated

        Parameters
        ---------
        station_id: int
            the station id
        channel_id: int
            the channel id
        zenith: float or None (default)
            the zenith angle of the incoming signal direction

        Returns string
        """
        antenna_type = self.get_antenna_type(station_id, channel_id)
        antenna_relative_position = self.get_relative_position(station_id, channel_id)

        antenna_model = ""
        if(zenith is not None and (antenna_type == 'createLPDA_100MHz')):
            if(antenna_relative_position[2] > 0):
                antenna_model = "{}_InfAir".format(antenna_type)
                if((not self.__assume_inf) and zenith < 90 * units.deg):
                    antenna_model = "{}_z1cm_InAir_RG".format(antenna_type)
            else:  # antenna in firn
                antenna_model = "{}_InfFirn".format(antenna_type)
                if((not self.__assume_inf) and zenith > 90 * units.deg):  # signal comes from below
                    antenna_model = "{}_z1cm_InFirn_RG".format(antenna_type)
                    # we need to add further distinction here
        elif(not antenna_type.startswith('analytic')):
            if(antenna_relative_position[2] > 0):
                antenna_model = "{}_InfAir".format(antenna_type)
            else:
                antenna_model = "{}_InfFirn".format(antenna_type)
        else:
            antenna_model = antenna_type
        return antenna_model

    def get_noise_RMS(self, station_id, channel_id, stage='amp'):
        """
        returns the noise RMS that was precomputed from forced triggers

        Parameters
        ----------
        station_id: int
            station id
        channel_id: int
            the channel id, not used at the moment, only station averages are computed
        stage: string (default 'amp')
            specifies the stage of reconstruction you want the noise RMS for,
            `stage` can be one of
             * 'raw' (raw measured trace)
             * 'amp' (after the amp was deconvolved)
             * 'filt' (after the trace was highpass with 100MHz

        Returns
        -------
        RMS: float
            the noise RMS (actually it is the standard deviation but as the mean should be zero its the same)
        """
        if(self.__noise_RMS is None):
            import json
            detector_directory = os.path.dirname(os.path.abspath(__file__))
            with open(os.path.join(detector_directory, 'noise_RMS.json'), 'r') as fin:
                self.__noise_RMS = json.load(fin)

        key = "{:d}".format(station_id)
        if(key not in self.__noise_RMS.keys()):
            rms = self.__noise_RMS['default'][stage]
            logger.warning("no RMS values for station {} available, returning default noise for stage {}: RMS={:.2g} mV".format(station_id, stage, rms / units.mV))
            return rms
        return self.__noise_RMS[key][stage]



import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
import os
from radiotools import helper as hp
from NuRadioReco.utilities import units, io_utilities
import logging
logger = logging.getLogger('analog_components')



def load_amp_response(amp_type='rno_surface', path=os.path.dirname(os.path.realpath(__file__))):  # use this function to read in log data
    """
    Read out amplifier gain and phase. Currently only examples have been implemented.
    Needs a better structure in the future, possibly with database.
    The hardware response incorporator currently reads in the load amp response.
    If you want to read in the RI function fur your reconstruction it needs to be changed
    in modules/RNO_G/hardweareResponseIncorporator.py l. 52, amp response.
    """
    amp_response = {}
    if amp_type == 'rno_surface':
        ph = os.path.join(path, 'HardwareResponses/surface_-60dBm_chan0_LogA_20dB.csv')
        ff = np.loadtxt(ph, delimiter=',', skiprows=7, usecols=0)
        amp_gain_discrete = np.loadtxt(ph, delimiter=',', skiprows=7, usecols= 5)
        amp_phase_discrete = np.loadtxt(ph, delimiter=',', skiprows=7, usecols= 6)
    elif amp_type == 'iglu':
        ph = os.path.join(path, 'HardwareResponses/iglu_drab_chan0_-60dBm_LogA_20dB.csv')
        ff = np.loadtxt(ph, delimiter=',', skiprows=7, usecols=0)
        amp_gain_discrete = np.loadtxt(ph, delimiter=',', skiprows=7, usecols= 5)
        amp_phase_discrete = np.loadtxt(ph, delimiter=',', skiprows=7, usecols= 6)
    else:
        logger.error("Amp type not recognized")
        return amp_response

    # Convert to GHz and add 20dB for attenuation in measurement circuit
    ff *= units.Hz

    if amp_type == 'rno_surface' or amp_type == 'iglu':
        amp_gain_discrete += 20
    amp_gain_db_f = interp1d(ff, amp_gain_discrete, bounds_error=False, fill_value=1)
    # all requests outside of measurement range are set to 0

    def get_amp_gain(ff):
        amp_gain_db = amp_gain_db_f(ff)
        amp_gain = 10 ** (amp_gain_db / 20.)  #convert from db to lin scale
        return amp_gain

    # Convert to MHz and broaden range
    amp_phase_f = interp1d(ff, np.unwrap(amp_phase_discrete * units.degree),
                           bounds_error=False, fill_value=0)  # all requests outside of measurement range are set to 0

    def get_amp_phase(ff):
        amp_phase = amp_phase_f(ff)
        return np.exp(1j * amp_phase)

    amp_response['gain'] = get_amp_gain
    amp_response['phase'] = get_amp_phase

    # def get_amp_response(ff):
    #    return amp_response['gain'](ff) * amp_response['phase'](ff)

    return amp_response

def load_amp_measurement(amp_measurement='surface_-60dBm_chan0_RI_20dB'):  # use this function to read in RI data
    """
    load individual amp measurement from file and buffer interpolation function
    """
    amp_measures = {}
    filename = os.path.join(os.path.dirname(__file__), 'HardwareResponses/', amp_measurement+".csv")
    ff = np.loadtxt(filename, delimiter=',', skiprows=7, usecols=0) # frequency
    ff *= units.Hz
    response_re = np.loadtxt(filename, delimiter=',', skiprows=7, usecols=5) # real part
    response_im = np.loadtxt(filename, delimiter=',', skiprows=7, usecols=6) # imaginary part
    response = response_re + response_im * 1j
    phase = np.unwrap(np.angle(response, deg=False))
    gain_db = (20 * np.log10(abs(response)))+20 #db
    gain = 10 ** (gain_db / 20.)

    amp_phase_f = interp1d(ff, phase, bounds_error=False, fill_value=0)  # all requests outside of measurement range are set to 0
    amp_gain_f = interp1d(ff, gain, bounds_error=False, fill_value=1)  # all requests outside of measurement range are set to 1

    amp_measures['gain'] = amp_gain_f
    amp_measures['phase'] = np.exp(1j * amp_phase_f)

    #def get_amp_measure(ff):
     #   return amp_measures['gain'](ff) * amp_measures['phase'](ff)

    return amp_measures


